#include "slideviewport.h"

#include "slideworld.h"
#include "sliderender.h"
#include "window.h"

BOOL CSLIDEViewport::m_bRenderViewportBound = FALSE;

//////////////////////////////////////////////////////////////////////////////
// CSLIDEViewport Class
//

CSLIDEViewport::CSLIDEViewport()
{
  m_pcWindowID = NULL;
}

CSLIDEViewport::~CSLIDEViewport()
{
}

VOID CSLIDEViewport::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			  CHAR *pcID, CHAR *pcWindowID)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  ASSERT( pcWindowID != NULL );

  m_pcWindowID = pcWindowID;
  m_vOrigin.Null();
  m_vSize.Set(1.0, 1.0, 0.0);

  m_listCSLIDERender.Init();
}

VOID CSLIDEViewport::Uninit()
{
  CSLIDEPrimitive::Uninit();
}

BOOL CSLIDEViewport::Initialized()
{
  if ( !CSLIDEPrimitive::Initialized() )
    {
      return FALSE;
    }

  return ( m_pcWindowID != NULL );
}

BOOL CSLIDEViewport::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iter;
  CWindow *pCWindow;

  if ( !CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Window Reference

  pCWindow = pCSLIDEWorld->PeekWindow(m_pcWindowID);
  if ( pCWindow == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtWindow],
		  m_pcWindowID);
      bCorrect = FALSE;
    }
  else
    {
      pCWindow->InsertViewport(this);
    }

  if ( !((0.0 <= m_vSize[X]) &&
	 (0.0 <= m_vSize[Y])) )
    {
      HandleError("%s \"%s\" has negative size (%.3f, %.3f)",
		  gcppcSLIDEReservedWords[sliderwtViewport],
		  m_pcID, m_vSize[X], m_vSize[Y]);
      bCorrect = FALSE;      
    }
  else
    {
      Vector vMax;
      vMax = m_vOrigin + m_vSize;
      if ( !( (0.0 <= m_vOrigin[X] && m_vOrigin[X] <= 1.0) &&
	      (0.0 <= m_vOrigin[Y] && m_vOrigin[Y] <= 1.0) &&
	      (0.0 <= vMax[X] && vMax[X] <= 1.0) &&
	      (0.0 <= vMax[Y] && vMax[Y] <= 1.0) ) )
	{
	  HandleError("%s \"%s\" does not lie entirely within its enclosing %s.\n(0, 0) < [ (%.3f, %.3f), (%.3f, %.3f) ] < (1, 1)", 
		      gcppcSLIDEReservedWords[sliderwtViewport],
		      m_pcID,
		      gcppcSLIDEReservedWords[sliderwtWindow],
		      m_vOrigin[X], m_vOrigin[Y], vMax[X], vMax[Y]);
	  bCorrect = FALSE;
	}
    }

  // Check Render Statement List

  if ( m_listCSLIDERender.GetLength() == 0 )
    {
      HandleWarning("%s \"%s\" has no %s statements associated with it", 
		    gcppcSLIDEReservedWords[sliderwtViewport], 
		    m_pcID, gcppcSLIDEReservedWords[sliderwtRender]);
    }
  else if ( m_listCSLIDERender.GetLength() > 1 )
    {
      HandleWarning("%s \"%s\" has %ld %s statements associated with it.\nThis may cause unpredictable results.", 
		    gcppcSLIDEReservedWords[sliderwtViewport], 
		    m_pcID, m_listCSLIDERender.GetLength(),
		    gcppcSLIDEReservedWords[sliderwtRender]);
    }

  return bCorrect;
}

BOOL CSLIDEViewport::Valid()
{
  Vector vMax;

  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  vMax = m_vOrigin + m_vSize;

  return ( (0.0 <= m_vOrigin[X] && m_vOrigin[X] <= 1.0) &&
	   (0.0 <= m_vOrigin[Y] && m_vOrigin[Y] <= 1.0) &&
	   (0.0 <= vMax[X] && vMax[X] <= 1.0) &&
	   (0.0 <= vMax[Y] && vMax[Y] <= 1.0) &&
	   (0.0 <= m_vSize[X] ) &&
	   (0.0 <= m_vSize[Y] ) );
}

VOID CSLIDEViewport::InsertRender(CSLIDERender *pCSLIDERender)
{
  ASSERT( pCSLIDERender != NULL );

  m_listCSLIDERender.InsertLast(pCSLIDERender, 0);
}

CSLIDEPrimitiveIter *CSLIDEViewport::AllocateIter()
{
  CSLIDEViewportIter *pCSLIDEViewportIter;

  pCSLIDEViewportIter = new CSLIDEViewportIter;
  ASSERT(pCSLIDEViewportIter != NULL );
  pCSLIDEViewportIter->Init(this);

  return pCSLIDEViewportIter;
}

VOID CSLIDEViewport::FreeIter(CSLIDEPrimitiveIter *pCSLIDEViewportIter)
{
  delete pCSLIDEViewportIter;
}

VOID CSLIDEViewport::CheckDynamic()
{
  SetDynamic(m_vOrigin.GetDynamic() || m_vSize.GetDynamic());
}

VOID CSLIDEViewport::UpdateState()
{
  // Resolve RenderID's if they are dynamic
  m_vOrigin.Update();
  m_vSize.Update();
}

VOID CSLIDEViewport::HandleReshape(UINT_32 uiWidth, UINT_32 uiHeight)
{
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;
  Vector vSize;

  vSize.Set(m_vSize[X]*uiWidth,
	    m_vSize[Y]*uiHeight,
	    0.0);

  // Pass the event down the render statements
  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)iter.PeekFirst(); 
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)iter.PeekNext() )
    {
      pCSLIDERender->HandleReshape(vSize);
    }
}

// Both Mouse Functions get the size of the window rectangle
// and the location of the mouse cursor relative to the lower 
// left hand corner
BOOL CSLIDEViewport::HandleMouseButton(UINT_32 uiWidth, UINT_32 uiHeight,
				       INT_32 iButton, INT_32 iState, 
				       INT_32 iX, INT_32 iY)
{
  BOOL b, bPostRedisplay;
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;
  Vector vClick, vSize;

  bPostRedisplay = FALSE;

  vClick.Set(iX - m_vOrigin[X]*uiWidth,
	     iY - m_vOrigin[Y]*uiHeight,
	     0.0);
  vSize.Set(m_vSize[X]*uiWidth,
	    m_vSize[Y]*uiHeight,
	    0.0);

  if ( iState == GLSI_DOWN )
    {
      // Check if the button down happened within our viewport
      // if not ignore
      if ( !((0.0 <= vClick[X] && vClick[X] <= vSize[X]) && 
	     (0.0 <= vClick[Y] && vClick[Y] <= vSize[Y])) )
	{
	  return FALSE;
	}
    }

  // Pass the event down the render statements
  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)iter.PeekFirst(); 
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)iter.PeekNext() )
    {
      b = pCSLIDERender->HandleMouseButton(vSize, iButton, iState, vClick);
      bPostRedisplay = bPostRedisplay || b;
    }

  return bPostRedisplay;
}

BOOL CSLIDEViewport::HandleMouseMotion(UINT_32 uiWidth, UINT_32 uiHeight,
				       INT_32 iX, INT_32 iY)
{
  BOOL b, bPostRedisplay;
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;
  Vector vClick, vSize;

  bPostRedisplay = FALSE;

  vClick.Set(iX - m_vOrigin[X]*uiWidth,
	     iY - m_vOrigin[Y]*uiHeight,
	     0.0);
  vSize.Set(m_vSize[X]*uiWidth,
	    m_vSize[Y]*uiHeight,
	    0.0);

  // Pass the event down the render statements
  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)iter.PeekFirst(); 
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)iter.PeekNext() )
    {
      b = pCSLIDERender->HandleMouseMotion(vSize, vClick);
      bPostRedisplay = bPostRedisplay || b;
    }

  return bPostRedisplay;
}

VOID CSLIDEViewport::Render(INT_32 iOriginX, INT_32 iOriginY,
			    UINT_32 uiWidth, UINT_32 uiHeight)
{
  Vector vOrigin, vSize;
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;

  vOrigin.Set(iOriginX + m_vOrigin[X]*uiWidth, 
	      iOriginY + m_vOrigin[Y]*uiHeight, 
	      0.0);
  vSize.Set(m_vSize[X]*uiWidth, 
	    m_vSize[Y]*uiHeight, 
	    1.0);
  
  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)iter.PeekFirst(); 
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)iter.PeekNext() )
    {
      pCSLIDERender->Render(vOrigin, vSize);
    }

  if ( GetRenderViewportBound() )
    {
      gilColor3f(1.0, 1.0, 0.0);
      gilBegin(GL_LINE_LOOP);
      {
        gilVertex2f(vOrigin[X], vOrigin[Y]);
        gilVertex2f(vOrigin[X] + vSize[X], vOrigin[Y]);
        gilVertex2f(vOrigin[X] + vSize[X], vOrigin[Y] + vSize[Y]);
        gilVertex2f(vOrigin[X], vOrigin[Y] + vSize[Y]);
      }
      gilEnd();
    }
}

VOID CSLIDEViewport::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtViewport], 
	  m_pcID, m_pcWindowID);
  fprintf(fpOutputFile, "  %s ( %.3f %.3f )\n", 
	  gcppcSLIDEReservedWords[sliderwtOrigin], 
	  m_vOrigin[X], m_vOrigin[Y]);
  fprintf(fpOutputFile, "  %s ( %.3f %.3f )\n", 
	  gcppcSLIDEReservedWords[sliderwtWinSize], 
	  m_vSize[X], m_vSize[Y]);
  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndViewport]);
}

UINT_32 CSLIDEViewport::GetRenderCount()
{
  return m_listCSLIDERender.GetLength();
}

//////////////////////////////////////////////////////////////////////////////
// CSLIDEViewportIter Class
//

CSLIDEViewportIter::CSLIDEViewportIter()
{
}

CSLIDEViewportIter::~CSLIDEViewportIter()
{
}

VOID CSLIDEViewportIter::Init(CSLIDEViewport *pCSLIDEViewport)
{
  CSLIDEPrimitiveIter::Init(pCSLIDEViewport);

  m_iter.Init(&(pCSLIDEViewport->m_listCSLIDERender));
}

VOID CSLIDEViewportIter::Uninit()
{
  CSLIDEPrimitiveIter::Uninit();
}

CSLIDERender *CSLIDEViewportIter::PeekFirstCSLIDERender()
{
  return (CSLIDERender*)(m_iter.PeekFirst());
}

CSLIDERender *CSLIDEViewportIter::PeekNextCSLIDERender()
{
  return (CSLIDERender*)(m_iter.PeekNext());
}
